@@ -85,7 +85,7 @@ def save_ongoing_games(data):
 
 def load_ongoing_games():
     debug("Trying to read ongoing games json")
-    data = None
+    data = {}
     try:
         with open(ONGOING_GAMES_STORE) as json_file:
             data = json.load(json_file)
@@ -973,7 +973,7 @@ def onEnter(nick, channel):
         game = botState.get_selected_game(nick, channel)
         msg.append(f"Game with {game.other(nick)}")
         msg += game.utf8_board(nick)
-        return msg
+        return Message(channel, message=msg)
 
 def load_game(nick, against_nick, channel, moves):
     debug(f"Loading {nick}")
